# default lib imports
import logging
import ssl
import sys
import os
import socket
import threading
from concurrent.futures import ThreadPoolExecutor

# external lib imports
import gevent

# module imports
from .Client import Client
from .Packet import Packet
from .Message import Message
from .DeviceType import DeviceType


# Manager class
class Manager:
    def __init__(self, ssl_context: ssl.SSLContext = None, host: str = "127.0.0.1", connection_port: int = 8595,
                 max_workers: int = 8, heartbeat_rate: int = 60, heartbeat_timeout: int = 10,
                 backlogged_connections: int = 10,
                 logging_id: str = "[Manager]", logging_level: int = logging.WARNING,
                 client_logging_level: int = logging.WARNING):

        # <> Instantiate Public Class Variables <>
        # the host is the computer running this program
        self.host = host

        # port that the manager will listen on for devices that want to connect
        self.port = connection_port

        # the delay between heartbeat messages
        if isinstance(heartbeat_rate, (int, float)):
            self.heartbeat_rate = heartbeat_rate
        else:
            raise TypeError("heartbeat_rate must be a number representing delay in seconds")

        # the time a client has to respond before being removed
        if isinstance(heartbeat_timeout, (int, float)):
            self.heartbeat_timeout = heartbeat_timeout
        else:
            raise TypeError("heartbeat_timeout must be a number representing time in seconds")

        # <> Instantiate Private Class Variables <>
        # socket that will be used for accepting new clients
        self.__connection_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # a pool of connected clients
        self.__client_pool = []

        # a list of device types
        self.__device_types = {}

        # logger object and setup
        if isinstance(logging_id, str):
            # logger object
            self.logger = logging.getLogger(logging_id)
        else:
            raise TypeError("logging_id must be of type str")

        if isinstance(logging_level, int):
            self.logger.setLevel(logging_level)
        else:
            raise TypeError("logging_level must resolve to an integer value")

        if isinstance(client_logging_level, int):
            self.client_logging_level = client_logging_level
        else:
            raise TypeError("client_logging_level must resolve to an integer value")

        # <> Server Socket Setup <>
        # true if the socket is using SSL
        self.__ssl_enabled = False

        # attempt to bind the socket to the provided host and port
        try:
            # try binding
            self.__connection_socket.bind((self.host, self.port))

        # in the case a socket error is raised
        except socket.error:
            raise Exception("Unable to bind server socket to the provided host and port")

        # attempt to set the socket as a listening socket
        try:
            # if the backlogged_connections is not type int then raise an exception
            if not isinstance(backlogged_connections, int):
                raise TypeError("backlogged_connections must be type int")

            # set the socket as a listening socket
            self.__connection_socket.listen(backlogged_connections)

        # in the case a socket error is raised
        except socket.error:
            raise Exception("Unable to set the socket as ")

        # if a SSL context is provided then try and wrap the socket
        if ssl_context is not None and isinstance(ssl_context, ssl.SSLContext):
            # try to use the SSL context provided
            try:
                # create a SSLSocket using the provided ssl_context
                self.__connection_socket = \
                    ssl_context.wrap_socket(self.__connection_socket, server_side=True)

                # set __ssl_enabled as true
                self.__ssl_enabled = True

                # logging output
                self.logger.info("(Constructor) Provided SSL context used to wrap socket successfully.")

            # in the case of a SSLError
            except ssl.SSLError as error:
                # logging output
                self.logger.error("(Constructor)  Unable to create SSL socket using the provided SSL context due to the"
                                  + " following SSLError: " + str(error) + ". Defaulting to the depreciated"
                                  + " ssl.wrap_socket() method. Do not use this in a production application.")

                # use the default ssl.wrap_socket method
                self.__connection_socket = ssl.wrap_socket(self.__connection_socket)

                # set __ssl_enabled as true
                self.__ssl_enabled = True

        # if the ssl_context provided
        elif ssl_context is not None and not isinstance(ssl_context, ssl.SSLContext):
            # logging output
            self.logger.error("(Constructor) ssl_context is not the proper type, please provide a sll.SSLContext."
                              + " Defaulting to the depreciated ssl.wrap_socket() method. Do not use this in a"
                              + " production application.")

            # use the default ssl.wrap_socket method
            self.__connection_socket = ssl.wrap_socket(self.__connection_socket)

            # set __ssl_enabled as true
            self.__ssl_enabled = True

        # <> Greenlet Setup and Initialization <>
        # a gevent ThreadPool which will be used when the manager needs to create temporary greenlets
        self.__executor = gevent.get_hub().threadpool

        # default thread pool used for async socket operations
        self.__thread_executor = ThreadPoolExecutor(max_workers=max_workers)

        # list of core threads
        self.__main_threads = []

    # run on a greenlet and is what connects new devices to the manager
    def __connection_listener(self):
        # log on init
        self.logger.debug("(Connection Listener) Starting process...")

        while True:
            try:
                # await a connection to the socket waiting for clients
                client_connection, client_address = self.__connection_socket.accept()

                # logging output
                self.logger.info("(Connection Listener) New Client Connection Established. Client IP Address: '"
                                 + str(client_address[0]) + "'")

                # hand over the client connection and address to a temporary greenlet which will get the data of the new
                # client
                self.__thread_executor.submit(self.__register_client, client_connection, client_address)

            # catch any SSL errors and let the user know about them so they can fix them on the client end
            except ssl.SSLError as exception:
                self.logger.error("(Connection Listener) Error occurred in the connection_listener greenlet."
                                  " Exception: " + str(exception))
            finally:
                # yield greenlet
                gevent.sleep(0.1)

    # takes a client socket and address and awaits the data which identifies the client's type
    def __register_client(self, connection: socket, address: tuple):
        try:
            # try to create the client while checking for errors
            try:
                # create the client object (will also retrieve the client's info)
                client = Client(connection, address, self.client_logging_level)

                # get the client's info
                client.get_data()

                # logging output
                self.logger.info("(Client Registrar) Got the data of client '" + client.uuid
                                 + "', adding client to pool.")

            # in the case the client gave an invalid type
            except Client.InvalidInfo:
                # logging output
                self.logger.error("(Client Registrar) Failed to receive valid data from client client with IP '"
                                  + str(address[0]) + "', ending connection.")

                # exit the function early
                return

            # check if a client is already connected which shares a UUID, assume that it is a invalid connection of the
            # client which is trying to reconnect and remove the old connection
            # go through all clients
            for index, check_client in enumerate(self.__client_pool):
                # check and see if the current client's UUID matches that of another client
                if check_client.uuid == client.uuid:
                    # logging output
                    self.logger.debug("(Client Registrar) Client with UUID '" + client.uuid + "' has overridden a"
                                      + " established connection. This was likely due to reconnection before a "
                                      + "heartbeat check but could also be caused by two clients sharing a UUID.")

                    # try to end the old client to end its connection if it still exists for some reason
                    try:
                        # send the checked client the end message
                        check_client.end()
                    finally:
                        # remove the client
                        self.__remove_client(check_client)

            # catch and exceptions within the general on_connect function
            try:
                # run the on_connect function for the client
                self.on_connect(client)

            # in case of a exception when executing handler
            except Exception as error:
                self.logger.error(
                    "(Client Registrar) Exception caught when running the on_connect() handler. Exception: '"
                    + str(error) + "'")

            # catch and exceptions within the device specific on_message function if any occur
            try:
                # check if the user provided a device specific handler for this client's device type, if so execute it
                if client.type in self.__device_types.keys():
                    self.__device_types[client.type].on_connect(client)

            # in case of a exception when executing handler
            except Exception as error:
                self.logger.error(
                    "(Client Registrar) Exception caught when running the on_connect() handler. For the " + client.type +
                    " client type. Exception: '" + str(error) + "'")

            # add the client to the pool
            self.__client_pool.append(client)

            # log the success
            self.logger.info("(Client Registrar) Client with UUID '" + client.uuid + "' successfully added to pool.")

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_tb.tb_lineno)
            self.logger.warning("EXCEPTION IN REGISTRAR: " + str(e))

    # heartbeat loop for checking if clients are still connected, run every x seconds based on the heartbeat_rate
    def __heartbeat_checker(self):
        # log on init
        self.logger.debug("(Heartbeat Checker) Starting process...")

        while True:
            # wait the delay for a heartbeat to be sent
            gevent.sleep(self.heartbeat_rate)

            # logging output
            self.logger.info("(Heartbeat Checker) Starting client heartbeat check. Number of clients: '"
                             + str(len(self.__client_pool)) + "'.")

            # check the heartbeat of the connected clients
            for index in range(len(self.__client_pool) - 1, -1, -1):
                # logging output
                self.logger.debug("(Heartbeat Checker) Checking the heartbeat of client '"
                                  + self.__client_pool[index].uuid + "'.")

                # submit the heartbeat check to the executor
                self.__thread_executor.submit(self.__heartbeat, index)

    # check the heartbeat of a client based on their given index
    def __heartbeat(self, client_index):
        client = self.__client_pool[client_index]

        # logging output
        self.logger.debug("(Heartbeat Checker for Client '" + client.uuid
                          + "') Initiating heartbeat check.")

        try:
            # check the heartbeat of the client
            client.heartbeat(self.heartbeat_timeout)

            # logging output
            self.logger.debug("(Heartbeat Checker for Client '" + client.uuid + "') Client passed heartbeat check.")

            return
        except Client.ConnectionEnd:
            try:
                # end the client's connection
                client.end()
            finally:
                # logging output
                self.logger.error("(Heartbeat Checker for Client '" + client.uuid
                                  + "') Client failed heartbeat check, removing from pool.")

                # remove the client
                self.__remove_client(client)

    # this loop checks for client messages to the server
    def __message_listener(self):
        # log on init
        self.logger.debug("(Message Listener) Starting process...")

        # loop for the lifetime of the program
        while True:
            # loop through each client
            for client in self.__client_pool:
                # check if the client has sent some data
                packet = client.recv(0)

                # if the client has sent no data move onto the next client
                if packet is None:
                    continue
                # if the client has sent some data then process it using the on_message function
                else:
                    # logging message
                    self.logger.debug("(Message Listener) Data received from client '" + client.uuid + "'."
                                      + " Sending data to appropriate handlers.")

                    # create a message
                    message = Message(client, packet)

                    # submit data to the __handle_message function
                    self.__executor.spawn(self.__handle_message(message))

            # yield
            gevent.sleep(0.1)

    # function which is used to handle data sent from the client
    def __handle_message(self, message: Message):
        # catch and exceptions within the general on_message function
        # noinspection PyBroadException
        try:
            # run the on_connect function for the client
            self.on_message(message)

        # in case of a exception when executing handler
        except Exception as error:
            self.logger.error("(Message Handler) Exception caught when running the on_message() handler. Exception: '"
                              + str(error) + "'")

        # catch and exceptions within the device specific on_message function if any occur
        # noinspection PyBroadException
        try:
            # check if the user provided a device specific handler for this client's device type, if so execute it
            if message.client.type in self.__device_types.keys():
                self.__device_types[message.client.type].on_message(message)

        # in case of a exception when executing handler
        except Exception as error:
            self.logger.error(
                "(Message Handler) Exception caught when running the on_message() handler for DeviceType '" +
                message.client.type + "." + "Exception: '" + str(error) + "'")

    # remove a client from the client pool and trigger the disconnect handler
    def __remove_client(self, client):
        # catch and exceptions within the general on_connect function
        try:
            # run the on_connect function for the client
            self.on_disconnect(client)

        # in case of a exception when executing handler
        except Exception as error:
            self.logger.error(
                "(Message Handler) Exception caught when running the on_disconnect() handler. Exception: '"
                + str(error) + "'")

        # catch and exceptions within the device specific on_message function if any occur
        try:
            # check if the user provided a device specific handler for this client's device type
            # if so execute it
            if client.type in self.__device_types.keys():
                self.__device_types[client.type].on_disconnect(client)

        # in case of a exception when executing handler
        except Exception as error:
            self.logger.error(
                "(Message Handler) Exception caught when running the on_disconnect() handler. For the " +
                client.type + " client type. Exception: '" + str(error) + "'")

        # remove item from list
        self.__client_pool = [x for x in self.__client_pool if x.instance_id != client.instance_id]

    # sends data to all clients
    def send_all(self, data):
        # for each client in the pool
        for client in self.__client_pool:
            # send the client the data
            self.__thread_executor.submit(client.send, data)

    # sends data to all clients of a specified device_type
    def send_type(self, device_type, packet: Packet):
        # for each client in the pool
        for client in self.__client_pool:
            # check if the client matches the provided type
            if client.type == device_type:
                # send the client the data
                self.__thread_executor.submit(client.send, packet)

    # sends data to a client given a uuid
    def send(self, unique_id: str, packet: Packet):
        # for each client in the pool
        for client in self.__client_pool:
            # check if the client matches the provided uuid
            if client.uuid == unique_id:
                # send the client the data
                self.__thread_executor.submit(client.send, packet)
                return

    # returns an list of dictionaries storing the connected clients' information
    def get_client_data(self):
        # data list
        info = []

        # loop through the clients and add their info to the list
        for client in self.__client_pool:
            # add the info for the given client
            info.append(client.return_data())

        # return the list of client data
        return info

    # event manager decorator function
    def event(self, coroutine):
        # handle general on_connect and on_message handlers
        if coroutine.__name__ == "on_connect" or coroutine.__name__ == "on_message":
            # logging output
            self.logger.info("(Event Handler) '" + coroutine.__name__ + "' handler was added successfully.")

            # replaces the existing coroutine with the provided one
            setattr(self, coroutine.__name__, coroutine)
            return True

    # used to add a device to the manager
    def add_device(self, device: DeviceType):
        # ensure the device is of type DeviceType
        if not isinstance(device, DeviceType):
            raise ValueError("device is not of type DeviceType")

        # ensure device.type is of type str
        if not isinstance(device.type, str):
            raise ValueError("device.type is not of type str")

        # add the device to the list of device types
        self.__device_types.update({device.type: device})

        # logging output
        self.logger.debug("(Device Adder) Successfully added DeviceType '" + device.type + "' to the manager.")

    # on connection function - run on client connection
    def on_connect(self, client: Client):
        return

    # on message function - runs when a client sends a message to the server
    def on_message(self, message: Message):
        return

    # on disconnect function - runs when the client disconnects from the server
    def on_disconnect(self, client: Client):
        return

    # runs all the handlers
    def __process_handler(self):
        # logging output
        self.logger.info("(Process Handler) Starting processes.")

        # add the core threads to the pool
        self.__main_threads.append(threading.Thread(target=self.__connection_listener, name="Connection Listener",
                                                    daemon=True))
        self.__main_threads.append(threading.Thread(target=self.__heartbeat_checker, name="Heartbeat Checker",
                                                    daemon=True))
        self.__executor.spawn(self.__message_listener)

        # start threads
        for thread in self.__main_threads:
            thread.start()

        # logging output
        self.logger.info("(Process Handler) All processes started.")
        return

    # starts all of the necessary permanent background threads
    def start(self):
        # start all of the background threads
        self.__process_handler()



# default lib imports
import threading
import json
import logging
from typing import Union
from queue import Queue
from uuid import uuid4

# external lib import
from gevent import socket

# module imports
from .ReservedBytes import GET_DATA, HEARTBEAT, END_CONNECTION
from .Packet import Packet


# Client class
class Client:
    # an exception raised when the client ends the connection, useful so the manager knows if a given client has
    # ended communications at any point
    class ConnectionEnd(Exception):
        pass

    # an exception raised when the client fails to get the client's info
    class InvalidInfo(Exception):
        pass

    def __init__(self, connection: socket, address: tuple, logging_level: int):
        # client logger
        self.logger = logging.getLogger("[Client](ID: UNKNOWN | IP: " + str(address[0]) + ")")
        self.logger_level = logging_level
        self.logger.setLevel(self.logger_level)

        # a lock preventing multiple threads from using the __connection at the same time
        self.connection_lock = threading.RLock()

        # socket object representing the connection between the client and the server
        self.__connection = connection

        # IP address of the client
        self.address = address

        # message buffer when irrelevant messages are received during a heartbeat
        self.buffer = Queue()

        # instance id of client (used to differentiate between multiple connections between the same client)
        self.instance_id = str(uuid4())

        # the uuid of the client
        self.uuid = None

        # the type of client the client is
        self.type = None

        # the bonus data which varies by client
        self.data = None

    # end the connection with the client
    def end(self, raise_exception: bool = False):
        """
        End the socket connection with the client.

        :param raise_exception: If a Client.ConnectionEnd exception should be raised when the connection is ended. True
                                will raise the exception False will end the connection quietly.

        :return: None
        """

        # logging output
        self.logger.warning("(end) Ending connection.")

        try:
            # tell the client the connection is ending
            self.send(Packet(END_CONNECTION))

        finally:
            # shutdown the socket connection
            self.__connection.shutdown(socket.SHUT_RDWR)

            # close the socket connection
            self.__connection.close()

            if raise_exception:
                # raise a connection end error
                raise self.ConnectionEnd

    # send data to the client
    def send(self, packet: Packet) -> bool:
        """
        Attempt to send a packet to the client.

        :param packet: Packet object containing the data to send to the client, should be a Packet in sending mode.

        :return: True if the packet was send, false if the packet could not be sent.
        """

        self.logger.debug("(send) Attempting to send a Packet.")

        # check if the packet object is a Packet in sending mode
        if not packet.is_sending():
            self.logger.warning("(send) Unable to send Packet, Packet object must be a sending Packet.")
            return False

        # init the lock
        with self.connection_lock:
            try:
                # send the string to to the client
                self.__connection.sendall(packet.bytes)

                self.logger.debug("(send) Packet sent successfully.")

                return True

            # in case of a socket error return None
            except socket.error:
                self.logger.debug("(send) Packet was unable to be sent.")
                return False

    # recv using the protocol of [2B - Message Size][Message] -> ignore rest of buffer
    def __smart_recv(self, timeout: int) -> Union[bytes, None]:
        """
        'Smartly' check the client socket's buffer.
        Read the first two bytes of the buffer (message header), this will give the size of the message.
        Read the # of bytes the message header specified and return those bytes.

        :param timeout: The time allowed for a response before the server assumes the client is dead and
                        ends the connection via the end() function.

        :return: Either None or a bytes object.
        """

        # init the lock
        with self.connection_lock:
            # if a timeout is specified then set the timeout
            if timeout is not None:
                self.__connection.settimeout(timeout)

            # read the first two bytes to get the size of the message
            message_size_bytes = self.__connection.recv(2)

            # if the buffer is empty return none
            if message_size_bytes is None:
                return None

            # if it has a value convert it into an integer
            message_size = int.from_bytes(message_size_bytes, 'big', signed=False)

            # log message header info
            self.logger.debug("(recv) Read message size header as: " + str(message_size_bytes) + " or int(" +
                              str(message_size) + ")")

            # recv up to the number of bytes we should recv
            message = self.__connection.recv(message_size)

            # log message body
            self.logger.debug("(recv) Read message body as: " + str(message))

            # check if the message len in bytes matches the length given, if it did not give a warning
            if message_size != len(message):
                # warn if the message length doesn't match the length given by the header
                self.logger.warning("(recv) Message length did not match the length specified in the message header.")

            return message

    # receives data from a client (optional timeout)
    def recv(self, timeout: int = 15) -> Union[Packet, None]:
        """
        Receives data from the client and returns either None if no new data has been received or a Packet
        object if the client has received some data.

        :param timeout: The time allowed for a response before the server assumes the client is dead and
                        ends the connection via the end() function.

        :return: None or a receiving Packet object.
        """

        # if there is a buffered message
        if self.buffer.qsize() > 0:
            return self.buffer.get()

        # blank response
        response = None

        # put the recv in a try catch to check for a timeout
        try:
            # read data from buffer
            response = self.__smart_recv(timeout)

        # in case of socket timeout
        except socket.timeout:
            return None

        # in case of another socket error
        except socket.error:
            return None

        finally:
            # if response is non return None
            if response is None:
                return None

            self.logger.debug("(recv) Returning received packet.")

            # return data as a packet object
            return Packet(response, sending=False)

    # gets the client's info (returns None if successful)
    def get_data(self, timeout: int = 15):
        """
        Get's the connected client's data when the client first connects to the manager. Will throw
        a Client.InvalidInfo exception if the client fails to provide a valid response or a Client.ConnectionEnd
        exception if the client fails to respond at all. (is called by the manager when the Client object is
        initialized but can be used to update the client if for some reason the Client's info changes in real
        time and must be updated on the server)

        :param timeout: The time allowed for a response before the server assumes the client is dead and
                        ends the connection via the end() function.

        :return: True if the connection is alive and False if the connection is now dead.
        """

        # logging output
        self.logger.info("(getinfo) Sending getinfo request.")

        # init the lock
        with self.connection_lock:
            # logging output
            self.logger.debug("(getinfo) Sending getinfo packet.")

            # ask the client for their data if the send fails then end the connection
            if not self.send(Packet(GET_DATA)):
                self.end(True)

            # logging output
            self.logger.debug("(getinfo) Receiving getinfo response")

            # await client response containing their info
            packet = self.recv(timeout=timeout)

            # if the response returned None because of an error end the connection
            if packet is None:
                self.end(True)

            # output the response
            self.logger.debug("(getinfo) Got data from client: '" + packet.bytes.decode() + "'.")

            # split the response using "##" as the delimiter
            response = packet.bytes.decode().split("##")

            # if the response is not 3 pieces of data raise an error
            if len(response) != 3:
                # logging output
                self.logger.error("(getinfo) Response invalid formatting, must be 3 values separated by delimiter '##'"
                                  + ". Formatted like: 'UUID##DEVICE_TYPE##JSON'.")

                raise self.InvalidInfo

            # get the clients UUID
            client_uuid = response[0]

            # update the client's logger info
            self.logger = logging.getLogger("[Client](ID: " + client_uuid + ")")
            self.logger.setLevel(self.logger_level)

            # check if the UUID is 36 char long to only allow proper UUIDs
            if len(client_uuid) != 36:
                # logging output
                self.logger.error("(getinfo) UUID invalid length, should be 36 characters. Data received: "
                                  + str(response) + ".")

                raise self.InvalidInfo

            # get the client's type
            client_type = response[1]

            # convert the data from a JSON string to a python dict
            try:
                # get the clients JSON data and convert it to a dict
                client_data = json.loads(response[2])
            except json.JSONDecodeError:
                # logging output
                self.logger.error("(getinfo) Failed to parse JSON.")

                # in case the JSON data provided is formatted incorrectly
                raise self.InvalidInfo

            # store the retrieved data
            self.uuid = client_uuid
            self.type = client_type
            self.data = client_data
            return None

    # check the heartbeat of the client returns None if client has a heartbeat
    def heartbeat(self, heartbeat_timeout: int) -> bool:
        """
        Checks the if the client socket is still alive by sending it a heartbeat packet, then waits for a response.
        If no response is received then the client is assumed to be dead and the socket is closed and the manager
        will remove the client and run the appropriate handler. (automatically done on a given interval by the manager
        but is available to users if they would like to manually initiate a heartbeat check)

        :param heartbeat_timeout: The time allowed for a response before the server assumes the client is dead and
                                  ends the connection via the end() function.

        :return: True if the connection is alive and False if the connection is now dead.
        """

        # logging output
        self.logger.info("(heartbeat) Sending heartbeat request.")

        # init the lock
        with self.connection_lock:
            # logging output
            self.logger.debug("(heartbeat) Sending heartbeat packet.")

            # send the heartbeat command if it fails end the connection
            if not self.send(Packet(HEARTBEAT)):
                self.end(True)

            # logging output
            self.logger.debug("(heartbeat) Receiving heartbeat response.")

            response = b''
            # loop until heartbeat or None is returned from recv
            while response is not None:
                # await a "beat" response or await timeout error
                response = self.recv(timeout=heartbeat_timeout)

                # if response is non exit the loop
                if response is None:
                    continue

                # log the bytes
                self.logger.debug("(heartbeat) Heartbeat response: " + str(response.bytes))

                # if the client's response is not beat then return a incorrect response
                if response.bytes == HEARTBEAT:
                    # logging output
                    self.logger.info("(heartbeat) Passed heartbeat check.")

                    # return the alive response
                    return True
                else:
                    self.logger.debug("(heartbeat) Response was not a heartbeat, adding response to buffer and " +
                                      "waiting again for heartbeat")

                    # put the received message into the message buffer
                    self.buffer.put(response)

            # if the recv fails then end the client connection and log a warning
            self.logger.warning("(heartbeat) Failed heartbeat check.")
            self.end(True)
            return False

    # client info as a dict
    def return_data(self) -> Union[dict, None]:
        """
        Returns the client's information as a JSON serializable dictionary. More reliable than dict() because it will
        return None if the client has not yet connected.

        :return: A dict object containing the client's data which is JSON serializable or None if the client's data
                 has not been retrieved yet.
        """

        # check if the data has not been returned
        if self.data is not None:
            return {
                "instance_id": self.instance_id,
                "uuid": self.uuid,
                "type": self.type,
                "data": self.data
            }
        else:
            return None

    # use dict(client) to get client info
    def __iter__(self):
        yield "instance_id", self.instance_id
        yield "uuid", self.uuid
        yield "type", self.type
        yield "data", self.data
